
package com.architech.connector;


/**
 * @author	Bjorn Stadheim (bjorn@metamerge.com)
 * @date	09.02.2001
 *
 * HTTPConnector provides a client side HTTP connection where the user can use it in
 * Iterator, Lookup and AddOnly mode. A parser may be specified in which case the contents
 * sent/received on the HTTP connection is parsed.
 *
 * If no parser is specified then all data is conveyed through the entry object. All HTTP
 * headers from the server is entered as is prefixing their names with "http.". The "body"
 * attribute contains the result from the server and also provides the data to be submitted
 * to the server if no parser is specified.
 *
 */
 
import com.architech.rscSearchCriteria;
import java.net.*;
import java.io.*;

import com.architech.entry.*;
import com.architech.parser.*;

public class HTTPConnector extends rscConnector {

	/**
	 * The connector name as reported in the log files
	 */
	private final String myName = "Metamerge HTTP Connector";

	/**
	 * The http object is the HttpURLConnection object obtained from the URL.
	 */
	private HttpURLConnection	http;

	/**
	 * Constructor
	 */
	public HTTPConnector () {
		setName (myName);
	}
	
	/**
	 * This function is called when the connector is no longer needed by the
	 * user (e.g. assembly line or script). Always call superclass' terminate which will
	 * take care of releasing resources, closing parsers etc.
	 */
	public void terminate () {

		super.terminate();
		
	}
	
	/**
	 * This function is called once after the connector configuration file has
	 * been provided by the caller.
	 * The single object parameter is usually null but can be any type of object
	 * the caller chooses to passon. Normally the parameter is some kind of input
	 * stream or Reader object.
	 */
	public void initialize (Object o) throws Exception {

	}
	
	/**
	 * This function is called when the connector operates in Iterator mode inside
	 * an assembly line.
	 */
	public void selectEntries() throws Exception {
	}
	
	/**
	 * This function is called to retreive the next entry from the connector. When
	 * there are no more entries to retrieve the function should return a null value
	 * indicating a logical end of file.
	 *
	 * If the connector has no parser this will never happen since we always make
	 * a HTTP connection every time getNextEntry is called.
	 *
	 */
	public Entry getNextEntry () throws Exception {
	
		initHttp ( getParam("url"), false );

		return result();
	}

	/**
	 * The putEntry function is called during AddOnly or Update operations.
	 */
	public void putEntry (Entry entry) throws Exception {
	
		initHttp ( getParam("url"), true );
	
		doOutput ( entry );
	}
	
	/**
	 * The findEntry function is called during Lookup or Update operations.
	 * Since this connector implements both findEntry and putEntry it is technically
	 * possible to use this connector in Update mode. However, the connector does not
	 * support the modEntry function so this will eventually fail.
	 */
	public Entry findEntry (rscSearchCriteria search) throws Exception {

		String urlStr = null;
		
		if ( search.size() == 1 && search.getFirstCriteriaName().equals("url") ) {
			urlStr = search.getFirstCriteriaValue ();
		} else {
			urlStr = getParam("url");
			urlStr += "?";
			for ( int i = 0; i < search.size(); i++ ) {
				urlStr += search.getCriteria(i).name;
				urlStr += "=";
				urlStr += search.getCriteria(i).value;
			}
		}
		
		debug ( "findEntry: URL = " + urlStr);		
		
		initHttp ( urlStr, false );
	
		return result();
	}
	
	/**
	 * doOutput sets the HTTP request headers and any data to be posted over the
	 * connection.
	 */
	private void doOutput ( Entry e ) throws Exception {

		boolean didOutput = false;
		BufferedReader input = null;
		BufferedWriter output;

		// Set HTTP request headers
		String[] names = e.getAttributeNames();
		for ( int i = 0; i < names.length; i++ ) {
			if ( names[i].startsWith ("http.") ) {
				http.setRequestProperty ( names[i].substring(5), e.getString( names[i] ) );
			}
		}

		names = e.getPropertyNames();
		for ( int i = 0; i < names.length; i++ ) {
			if ( names[i].startsWith ("http.") ) {
				http.setRequestProperty ( names[i].substring(5), e.getProperty( names[i] ).toString() );
			}
		}
		
		// Create a buffered writer for the output stream
		// NOTE! Once the outputstream has been obtained, the http object will no longer
		// accept any changes to headers etc .
		output = new BufferedWriter ( new OutputStreamWriter ( http.getOutputStream() ) );	

		// If we have a parser use that one		
		if ( hasConfigValue("parser")  ) {
		
			debug ("Initialize parser with socket output stream");

			CharArrayWriter caw = new CharArrayWriter();

			initParser (null, caw);

			getParser().writeEntry ( e );
			getParser().closeParser ( );
			
			caw.close();
			caw.writeTo ( output );
			
		} else {
		
			debug ( "Get http.body from attribute/property http.body");

			// Get value object for the data attribute
			Object data = e.getAttribute ("http.body");
			if ( data != null )
				data = ((Attribute)data).getValue(0);

			// Try property if no attribute
			if ( data == null )
				data = e.getProperty ("http.body");
				
				
			// Is the provided data value an InputStream or Reader?
			if ( data instanceof InputStream || data instanceof Reader) {

				if ( data instanceof InputStream )
					input = new BufferedReader ( new InputStreamReader ((InputStream) data) );
				else
					input = new BufferedReader ( (Reader) data );
				
			
			// ... else dump the String representation
			} else {
				if ( data != null )
					input = new BufferedReader ( new StringReader ( data.toString() ) );
					//output.write ( data.toString() );
			}
			
			if ( input != null && getParser() == null ) {			
				debug ("Copy contents of " + data.getClass().getName() + " to HTTP connection");
				char[]	buffer = new char[4096];
				int count;
				
				while ( (count = input.read (buffer)) > 0 ) {
					output.write ( buffer, 0, count );
				}
			}
			
		}

		// Flush and close output writer
		output.flush ();
		output.close ();

		// The following calls adds two attributes to the passed entry (which in assembly line
		// context is the "conn" object).
		
		Entry res = result();
		
		e.merge ( res );
		//e.setAttribute ( "http.responseCode", new Integer (http.getResponseCode()) );
		//e.setAttribute ( "http.responseMsg", http.getResponseMessage() );
	}

	/**
	 * This function builds an entry with the HTTP headers and any data sent
	 * over the connection. If the content type is of type "text/" then we provide
	 * an additional attribute "body" that holds the text of the document.
	 */
	private Entry result () throws Exception {

		Entry e = new Entry();
		
		e.setAttribute ( "http.responseCode", new Integer (http.getResponseCode()) );
		e.setAttribute ( "http.responseMsg", http.getResponseMessage() );
		e.setAttribute ( "http.content-type", http.getContentType() );
		e.setAttribute ( "http.content-encoding", http.getContentEncoding() );
		e.setAttribute ( "http.content-length", new Integer (http.getContentLength()) );
		e.setAttribute ( "http.body", http.getContent() );
		
		if ( http.getContentType().startsWith("text/") ) {
			StringBuffer buffer = new StringBuffer ();
			InputStream is = (InputStream) http.getContent();
			for ( int i = 0; i < http.getContentLength(); i++ )
				buffer.append ( (char) is.read() );
		
			e.setAttribute ( "http.text-body", buffer );
		}
		
		//http.connect ();
		
		return e;
	}

	/**
	 * This function sets up the HTTP connection. If we have a parser then this
	 * is initialized with the connection input/output streams.
	 */
	private void initHttp (String strURL, boolean output) throws Exception {

		if ( strURL == null )
			throw new com.architech.exceptions.MissingConfigurationException ("HTTPConnector", "url");
		
		debug ("connect to: " + strURL);
		
		URL url = new URL ( strURL );
		http = (HttpURLConnection) url.openConnection();
		http.setDoOutput ( output );
		http.setRequestMethod ( (String) getParam("method") );

		/*		
		if ( getParser() != null ) {
			if ( output )
				initParser (null, http.getOutputStream());
			else
				initParser (http.getInputStream(), null);
		}
		*/
	}
}
